import { Role } from './ad.types';
import { Target } from './candidate.types';
import { CarpoolPathCreatorException } from './match.errors';
import { Actor } from './value-objects/actor.value-object';
import { Point } from './value-objects/point.value-object';
import { CarpoolStep } from './value-objects/carpool-step.value-object';

export class CarpoolPathCreator {
  private PRECISION = 5;

  constructor(
    private readonly driverWaypoints: Point[],
    private readonly passengerWaypoints: Point[],
  ) {
    if (driverWaypoints.length < 2)
      throw new CarpoolPathCreatorException(
        new Error('At least 2 driver waypoints must be defined'),
      );
    if (passengerWaypoints.length < 2)
      throw new CarpoolPathCreatorException(
        new Error('At least 2 passenger waypoints must be defined'),
      );
  }

  /**
   * Creates a path (a list of carpoolSteps) between driver waypoints
      and passenger waypoints respecting the order
      of the driver waypoints
      Inspired by :
      https://stackoverflow.com/questions/849211/shortest-distance-between-a-point-and-a-line-segment
   */
  public carpoolPath = (): CarpoolStep[] =>
    this._consolidate(
      this._mixedCarpoolSteps(
        this._driverCarpoolSteps(),
        this._passengerCarpoolSteps(),
      ),
    );

  private _mixedCarpoolSteps = (
    driverCarpoolSteps: CarpoolStep[],
    passengerCarpoolSteps: CarpoolStep[],
  ): CarpoolStep[] =>
    driverCarpoolSteps.length == 2
      ? this._simpleMixedCarpoolSteps(driverCarpoolSteps, passengerCarpoolSteps)
      : this._complexMixedCarpoolSteps(
          driverCarpoolSteps,
          passengerCarpoolSteps,
        );

  private _driverCarpoolSteps = (): CarpoolStep[] =>
    this.driverWaypoints.map(
      (waypoint: Point, index: number) =>
        new CarpoolStep({
          point: new Point({
            lon: waypoint.lon,
            lat: waypoint.lat,
          }),
          actors: [
            new Actor({
              role: Role.DRIVER,
              target: this._getTarget(index, this.driverWaypoints),
            }),
          ],
        }),
    );

  /**
   * Creates the passenger carpoolSteps with original passenger waypoints, adding driver waypoints that are the same
   */
  private _passengerCarpoolSteps = (): CarpoolStep[] => {
    const carpoolSteps: CarpoolStep[] = [];
    this.passengerWaypoints.forEach(
      (passengerWaypoint: Point, index: number) => {
        const carpoolStep: CarpoolStep = new CarpoolStep({
          point: new Point({
            lon: passengerWaypoint.lon,
            lat: passengerWaypoint.lat,
          }),
          actors: [
            new Actor({
              role: Role.PASSENGER,
              target: this._getTarget(index, this.passengerWaypoints),
            }),
          ],
        });
        if (
          this.driverWaypoints.filter((driverWaypoint: Point) =>
            passengerWaypoint.isSame(driverWaypoint),
          ).length == 0
        ) {
          carpoolStep.actors.push(
            new Actor({
              role: Role.DRIVER,
              target: Target.NEUTRAL,
            }),
          );
        }
        carpoolSteps.push(carpoolStep);
      },
    );
    return carpoolSteps;
  };

  private _simpleMixedCarpoolSteps = (
    driverCarpoolSteps: CarpoolStep[],
    passengerCarpoolSteps: CarpoolStep[],
  ): CarpoolStep[] => [
    driverCarpoolSteps[0],
    ...passengerCarpoolSteps,
    driverCarpoolSteps[1],
  ];

  private _complexMixedCarpoolSteps = (
    driverCarpoolSteps: CarpoolStep[],
    passengerCarpoolSteps: CarpoolStep[],
  ): CarpoolStep[] => {
    let mixedCarpoolSteps: CarpoolStep[] = [...driverCarpoolSteps];
    const originInsertIndex: number = this._insertIndex(
      passengerCarpoolSteps[0],
      driverCarpoolSteps,
    );
    mixedCarpoolSteps = [
      ...mixedCarpoolSteps.slice(0, originInsertIndex),
      passengerCarpoolSteps[0],
      ...mixedCarpoolSteps.slice(originInsertIndex),
    ];
    const destinationInsertIndex: number =
      this._insertIndex(
        passengerCarpoolSteps[passengerCarpoolSteps.length - 1],
        driverCarpoolSteps,
      ) + 1;
    mixedCarpoolSteps = [
      ...mixedCarpoolSteps.slice(0, destinationInsertIndex),
      passengerCarpoolSteps[passengerCarpoolSteps.length - 1],
      ...mixedCarpoolSteps.slice(destinationInsertIndex),
    ];
    return mixedCarpoolSteps;
  };

  private _insertIndex = (
    targetCarpoolStep: CarpoolStep,
    carpoolSteps: CarpoolStep[],
  ): number =>
    this._closestSegmentIndex(targetCarpoolStep, this._segments(carpoolSteps)) +
    1;

  private _segments = (carpoolSteps: CarpoolStep[]): CarpoolStep[][] => {
    const segments: CarpoolStep[][] = [];
    carpoolSteps.forEach((carpoolStep: CarpoolStep, index: number) => {
      if (index < carpoolSteps.length - 1)
        segments.push([carpoolStep, carpoolSteps[index + 1]]);
    });
    return segments;
  };

  private _closestSegmentIndex = (
    carpoolStep: CarpoolStep,
    segments: CarpoolStep[][],
  ): number => {
    const distances: Map<number, number> = new Map();
    segments.forEach((segment: CarpoolStep[], index: number) => {
      distances.set(index, this._distanceToSegment(carpoolStep, segment));
    });
    const sortedDistances: Map<number, number> = new Map(
      [...distances.entries()].sort((a, b) => a[1] - b[1]),
    );
    const [closestSegmentIndex] = sortedDistances.keys();
    return closestSegmentIndex;
  };

  private _distanceToSegment = (
    carpoolStep: CarpoolStep,
    segment: CarpoolStep[],
  ): number =>
    parseFloat(
      Math.sqrt(this._distanceToSegmentSquared(carpoolStep, segment)).toFixed(
        this.PRECISION,
      ),
    );

  private _distanceToSegmentSquared = (
    carpoolStep: CarpoolStep,
    segment: CarpoolStep[],
  ): number => {
    const length2: number = this._distanceSquared(
      segment[0].point,
      segment[1].point,
    );
    if (length2 == 0)
      return this._distanceSquared(carpoolStep.point, segment[0].point);
    const length: number = Math.max(
      0,
      Math.min(
        1,
        ((carpoolStep.point.lon - segment[0].point.lon) *
          (segment[1].point.lon - segment[0].point.lon) +
          (carpoolStep.point.lat - segment[0].point.lat) *
            (segment[1].point.lat - segment[0].point.lat)) /
          length2,
      ),
    );
    const newPoint: Point = new Point({
      lon:
        segment[0].point.lon +
        length * (segment[1].point.lon - segment[0].point.lon),
      lat:
        segment[0].point.lat +
        length * (segment[1].point.lat - segment[0].point.lat),
    });
    return this._distanceSquared(carpoolStep.point, newPoint);
  };

  private _distanceSquared = (point1: Point, point2: Point): number =>
    parseFloat(
      (
        Math.pow(point1.lon - point2.lon, 2) +
        Math.pow(point1.lat - point2.lat, 2)
      ).toFixed(this.PRECISION),
    );

  private _getTarget = (index: number, waypoints: Point[]): Target =>
    index == 0
      ? Target.START
      : index == waypoints.length - 1
      ? Target.FINISH
      : Target.INTERMEDIATE;

  /**
   * Consolidate carpoolSteps by removing duplicate actors (eg. driver with neutral and start or finish target)
   */
  private _consolidate = (carpoolSteps: CarpoolStep[]): CarpoolStep[] => {
    const uniquePoints: Point[] = [];
    carpoolSteps.forEach((carpoolStep: CarpoolStep) => {
      if (
        uniquePoints.find((point: Point) => point.isSame(carpoolStep.point)) ===
        undefined
      )
        uniquePoints.push(
          new Point({
            lon: carpoolStep.point.lon,
            lat: carpoolStep.point.lat,
          }),
        );
    });
    return uniquePoints.map(
      (point: Point) =>
        new CarpoolStep({
          point,
          actors: carpoolSteps
            .filter((carpoolStep: CarpoolStep) =>
              carpoolStep.point.isSame(point),
            )
            .map((carpoolStep: CarpoolStep) => carpoolStep.actors)
            .flat(),
        }),
    );
  };
}
